#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <string.h>
#include <getopt.h>

#include <gdk/gdkx.h>
#include <X11/Xlib.h>
#include <gtk/gtk.h>

#include <gksuui.h>
#include <gksu.h>

#include "defines.h"
#include "../config.h"

#include "util.h"

struct option long_opts[] = {
    /*
     * { name  has_arg  *flag  val } 
     */
    {"help", no_argument, NULL, 'h'},
    {"login", no_argument, NULL, 'l'},
    {"preserv-env", no_argument, NULL, 'k'},
    {"preserve-env", no_argument, NULL, 'k'},
    {"user", required_argument, NULL, 'u'},
    {"print-pass", no_argument, NULL, 'p'},
    {"message", required_argument, NULL, 'm'},
    {"title", required_argument, NULL, 't'},
    {"icon", required_argument, NULL, 'i'},
    {"disable-grab", no_argument, NULL, 'g'},
    {"ssh-fwd", no_argument, NULL, 's'},
    {"debug", no_argument, NULL, 'd'},
    {0, 0, 0, 0}
};

/**
 * help:
 * @cmdname:  name of the command which was called by the user 
 * (argv[0])
 *
 * This function is a simple 'usage'-style printing function.
 * It is called if the user calls the program with --help or -h
 */
void
help (gchar *cmdname)
{ 
  g_print (_(
"GKsu version %s\n\n"
"Usage: %s [-u <user>] [-k] [-l] <command>\n\n"
"  --user <user>, -u <user>\n"
"    Calls <command> as the specified user\n"
"  --message <message>, -m <message>\n"
"    Replaces the standard message shown to ask for\n"
"    password for the argument passed to the option\n"
"  --title <title>, -t <title>\n"
"    Replaces the default title with the argument\n"
"  --icon <icon>, -i <icon>\n"
"    Replaces the default window icon with the argument\n"
"  --print-pass, -p\n"
"    Asks gksu to print the password to stdout, just\n"
"    like ssh-askpass. Useful to use in scripts with\n"
"    programs that accept receiving the password on\n"
"    stdin.\n"
"  --disable-grab, -g\n"
"    Disables the \"locking\" of the keyboard, mouse,\n"
"    and focus done by the program when asking for\n"
"    password\n"
"  --ssh-fwd, -s\n"
"    Strip the host part of the $DISPLAY variable, so that\n"
"    GKSu will work on SSH X11 Forwarding.\n"
"\n" 
"  --login, -l\n"
"    Makes this a login shell. Beware this may cause\n"
"    problems with the Xauthority magic. Run xhost\n"
"    to allow the target user to open windows on your\n"
"    display!\n"
"  --preserve-env, -k\n"
"    Preserve the current environments, does not set $HOME\n"
"    nor $PATH, for example.\n"
"\n"
"\n"), PACKAGE_VERSION, cmdname);
}

/* copied from gnome-ssh-askpass */
#define GRAB_TRIES	16
#define GRAB_WAIT	250 /* milliseconds */

typedef enum
  {
    FAILED_GRAB_MOUSE,
    FAILED_GRAB_KEYBOARD
  } FailedGrabWhat;

void
report_failed_grab (FailedGrabWhat what)
{
  switch (what)
    {
    case FAILED_GRAB_MOUSE:
      gk_dialog (GTK_MESSAGE_WARNING, 
	     _("Could not grab your mouse.\n"
	       "A malicious client may be eavesdropping\n"
	       "on your session."));
      break;
    case FAILED_GRAB_KEYBOARD:
      gk_dialog (GTK_MESSAGE_WARNING, 
	     _("Could not grab your keyboard.\n"
	       "A malicious client may be eavesdropping\n"
	       "on your session."));
      break;
    }
}

void
grab_keyboard_and_mouse (GtkWidget *dialog)
{
  GdkGrabStatus status;
  gint grab_tries = 0;
  
  gtk_widget_show_all (dialog);

  while (gtk_events_pending())
    gtk_main_iteration ();

  for(;;) 
    {
      status = gdk_pointer_grab((GTK_WIDGET(dialog))->window, TRUE, 0, NULL,
				NULL, GDK_CURRENT_TIME);
      if (status == GDK_GRAB_SUCCESS)
	break;
      usleep(GRAB_WAIT * 1000);
      if (++grab_tries > GRAB_TRIES)
	{
	  report_failed_grab (FAILED_GRAB_MOUSE);
	  break;
	}
    }
  
  for(;;) 
    {
      status = gdk_keyboard_grab((GTK_WIDGET(dialog))->window,
				 FALSE, GDK_CURRENT_TIME);
      if (status == GDK_GRAB_SUCCESS)
	break;
      
      usleep(GRAB_WAIT * 1000);
      
      if (++grab_tries > GRAB_TRIES) 
	{
	  report_failed_grab (FAILED_GRAB_KEYBOARD);
	  break;
	}
    }
  
  gdk_x11_grab_server();
}

void
ungrab_keyboard_and_mouse ()
{
  /* Ungrab */
  XUngrabServer(GDK_DISPLAY());
  gdk_pointer_ungrab(GDK_CURRENT_TIME);
  gdk_keyboard_ungrab(GDK_CURRENT_TIME);
  gdk_flush();
}

int
main (int argc, char **argv)
{
  GtkWidget *dialog;
  GksuContext *context;

  gboolean print_pass = FALSE;
  gboolean grab = TRUE;
  gboolean sudo_mode = FALSE;

  gboolean message_changed = FALSE;

  gchar *password = NULL;
  GError *error = NULL;

  gint newargc = 0;
  gchar **newargv = NULL;

  int retvalue = 0;
  int c = 0;

  setlocale (LC_ALL, "");
  bindtextdomain(PACKAGE_NAME, LOCALEDIR);  
  bind_textdomain_codeset (PACKAGE_NAME, "UTF-8");
  textdomain(PACKAGE_NAME);

  { /* support gksu_sudo_run */
    gchar *myname = g_path_get_basename (argv[0]);
    if (!strcmp(myname, "gksudo"))
      sudo_mode = TRUE;
    g_free (myname);
  }

  /* 
   * bad, bad code... adds a second -- right after the first one,
   * because gtk_init will remove one of them... 
   */
  {
    /* to check whether a -- was already found when parsing arguments */
    gboolean separator_found = 0;

    for (c = 0; c < argc; c++)
      {
	if (!strcmp ("--", argv[c]) && (!separator_found))
	  {
	    newargv = g_realloc (newargv, sizeof(char*) * (newargc + 2));
	    newargv[newargc] = g_strdup (argv[c]);
	    newargv[newargc + 1] = g_strdup (argv[c]);
	    
	    newargc = newargc + 2;
	    separator_found = TRUE;
	  }
	else
	  {
	    newargv = g_realloc (newargv, sizeof(char*) * (newargc + 1));
	    newargv[newargc] = g_strdup (argv[c]);
	    
	    newargc++;
	  }
      }
  }

  gtk_init (&newargc, &newargv);

  dialog = gksuui_dialog_new ();

  context = gksu_context_new ();
  while ((c = getopt_long(newargc, newargv, "?hu:lpm:kt:i:gds", long_opts, NULL))
	 != EOF)
    {
      switch (c)
	{
	case 0:
	  break;

	case 'h':
	  help (newargv[0]);
	  exit(0);
	  break;
	case '?':
	  help (newargv[0]);
	  exit(0);
	  break;
	case 'u':
	  gksu_context_set_user (context, optarg);
	  break;
	case 'l':
	  gksu_context_set_login_shell (context, TRUE);
	  break;
	case 'p':
	  print_pass = TRUE;
	  break;
	case 't':
	  gtk_window_set_title (GTK_WINDOW(dialog), optarg);
	  break;
	case 'm':
	  gksuui_dialog_set_message (GKSUUI_DIALOG(dialog), optarg);
	  message_changed = TRUE;
	  break;
	case 'i':
	  {
	    GdkPixbuf *pixbuf;
	    pixbuf = gdk_pixbuf_new_from_file (optarg, NULL);
	    if (pixbuf)
	      gksuui_dialog_set_icon (GKSUUI_DIALOG(dialog), pixbuf);
	  }
	  break;
	case 'k':
	  gksu_context_set_keep_env (context, TRUE);
	  break;
	case 'g':
	  grab = FALSE;
	  break;
	case 'd':
	  gksu_context_set_debug (context, TRUE);
	  break;
	case 's':
	  gksu_context_set_ssh_fwd (context, TRUE);
	  break;
	}
    }

  if (print_pass)
    {
      if (!gksuui_dialog_get_message (GKSUUI_DIALOG(dialog)))
	{
	  gchar *msg = 
	    g_strdup_printf (_("<b>Please enter %s's password</b>"),
			     gksu_context_get_user (context));

	  gksuui_dialog_set_message (GKSUUI_DIALOG(dialog), msg);
	  g_free (msg);
	}

      if (grab)
	grab_keyboard_and_mouse (dialog);
      retvalue = gtk_dialog_run (GTK_DIALOG(dialog));
      gtk_widget_hide (dialog);
      if (grab)
	ungrab_keyboard_and_mouse ();

      /* 
	 the user may have pressed cancel or
	 closed the window 
      */
      if (retvalue != GTK_RESPONSE_OK)
	  return 0;

      {
	gchar *tmp;
	tmp = gksuui_dialog_get_password (GKSUUI_DIALOG(dialog));
	password = g_locale_from_utf8 (tmp, strlen (tmp), NULL, NULL, NULL);
	g_free (tmp);
      }

      if (password)
	printf ("%s\n", password);
      return 0;
    }

  /* now we can begin to care about a command */
  if (newargc <= optind)
    {
      gk_dialog (GTK_MESSAGE_ERROR, _("Missing command to run."));
      return 1;
    }

  {
    gchar *command = g_strdup (newargv[optind]);
    gchar *tmp = NULL;
    gint i = 0;

    for (i = optind + 1; i < newargc; i++)
      {
  	tmp = g_strconcat (command, " '", newargv[i], "'", NULL);
	g_free (command);
	command = tmp;
      }
    gksu_context_set_command (context, command);
    g_free (command);
  }

  {
    struct passwd *pwentry;

    pwentry = getpwnam (gksu_context_get_user (context));

    if (pwentry->pw_uid == geteuid ())
      return g_spawn_command_line_sync (gksu_context_get_command (context),
					NULL, NULL, NULL, NULL);
  }


  if (sudo_mode)
    {
      /* test if we can run without a password */
      gksu_context_sudo_run (context, &error);
      if (error)
	{
	  if (!message_changed)
	    {
	      gchar *msg = 
		g_strdup_printf (_("<b>Please enter your password\n"
				   "to run %s as user %s</b>"),
				 gksu_context_get_command (context),
				 gksu_context_get_user (context));
	      
	      gksuui_dialog_set_message (GKSUUI_DIALOG(dialog), msg);
	      g_free (msg);
	    }

	  if (grab)
	    grab_keyboard_and_mouse (dialog);
	  retvalue = gtk_dialog_run (GTK_DIALOG(dialog));
	  gtk_widget_hide (dialog);
	  if (grab)
	    ungrab_keyboard_and_mouse ();

	  while (gtk_events_pending ())
	    gtk_main_iteration ();

	  if (retvalue != GTK_RESPONSE_OK)
	    return 0;

	  {
	    gchar *tmp;
	    tmp = gksuui_dialog_get_password (GKSUUI_DIALOG(dialog));
	    password = g_locale_from_utf8 (tmp, strlen (tmp), NULL, NULL, NULL);
	    g_free (tmp);
	  }

	  gksu_context_set_password (context, password);
      
	  error = NULL;
	  gksu_context_sudo_run (context, &error);
	  if (error)
	    gk_dialog (GTK_MESSAGE_ERROR, 
		       _("Failed to run %s as user %s:\n %s"),
		       gksu_context_get_command (context),
		       gksu_context_get_user (context),
		       error->message);
	}
    }
  else
    {
      if (!message_changed)
	{
	  gchar *msg = 
	    g_strdup_printf (_("<b>Please enter %s's password\n"
			       "to run %s.</b>"),
			     gksu_context_get_user (context),
			     gksu_context_get_command (context));
	  
	  gksuui_dialog_set_message (GKSUUI_DIALOG(dialog), msg);
	  g_free (msg);
	}

      if (grab)
	grab_keyboard_and_mouse (dialog);
      retvalue = gtk_dialog_run (GTK_DIALOG(dialog));
      gtk_widget_hide (dialog);
      if (grab)
	ungrab_keyboard_and_mouse ();

      while (gtk_events_pending ())
	gtk_main_iteration ();

      if (retvalue != GTK_RESPONSE_OK)
	  return 0;

      {
	gchar *tmp;
	tmp = gksuui_dialog_get_password (GKSUUI_DIALOG(dialog));
	password = g_locale_from_utf8 (tmp, strlen (tmp), NULL, NULL, NULL);
	g_free (tmp);
      }

      gksu_context_set_password (context, password);
      gksu_context_run (context, &error);

      if (error)
	gk_dialog (GTK_MESSAGE_ERROR, 
		   _("Failed to run %s as user %s:\n %s"),
		   gksu_context_get_command (context),
		   gksu_context_get_user (context),
		   error->message);
    }
  
  return 0;
}
